import datetime
import enum
import typing

import dataclasses as da
import stringcase

UNKNOWN = 'UNKNOWN'


class UnknowableEnum(enum.Enum):
    @classmethod
    def _missing_(cls, value):
        return cls.unknown


class TableEnum(UnknowableEnum):
    unknown = UNKNOWN
    operlog = 'OPERLOG'
    attlog = 'ATTLOG'
    attphoto = 'ATTPHOTO'


class AlarmEnum(UnknowableEnum):
    unknown = UNKNOWN
    door_close_detected = '50'
    door_open_detected = '51'
    machine_been_broken = '55'
    out_door_button = '53'
    door_broken_accidentally = '54'
    try_invalid_verification = '58'
    alarm_cancelled = '65535'


class OperationEnum(UnknowableEnum):
    unknown = UNKNOWN
    start_up = '0'
    shutdown = '1'
    validation_failure = '2'
    alarm = '3'
    enter_the_menu = '4'
    change_settings = '5'
    registration_fingerprint = '6'
    registration_password = '7'
    card_registration = '8'
    delete_user = '9'
    delete_fingerprints = '10'
    delete_the_password = '11'
    delete_rf_card = '12'
    remove_data = '13'
    mf_create_cards = '14'
    mf_registration_cards = '15'
    mf_registration_cards_2 = '16'
    mf_registration_card_deleted = '17'
    mf_clearance_card_content = '18'
    moved_to_the_registration_card_data = '19'
    the_data_in_the_card_copied_to_the_machine = '20'
    set_time = '21'
    restore_factory_settings = '22'
    delete_records_access = '23'
    remove_administrator_rights = '24'
    group_set_up_to_amend_access = '25'
    modify_user_access_control_settings = '26'
    access_time_to_amend_paragraph = '27'
    amend_unlock_portfolio = '28'
    unlock = '29'
    registration_of_new_users = '30'
    fingerprint_attribute_changes = '31'
    stress_alarm = '32'


@da.dataclass(frozen=True)
class ServerDatetimeMixin:
    server_datetime: typing.Optional[datetime.datetime]

    def correct_datetime(
            self, tz: datetime.timezone
    ) -> typing.Optional[datetime.datetime]:
        if not self.server_datetime:
            return None
        return datetime.datetime(
            year=self.server_datetime.year,
            month=self.server_datetime.month,
            day=self.server_datetime.day,
            hour=self.server_datetime.hour,
            minute=self.server_datetime.minute,
            second=self.server_datetime.second,
            tzinfo=tz,
        )


@da.dataclass(frozen=True)
class Transaction(ServerDatetimeMixin):
    pin: str
    raw: str
    check_type: str = ''
    verify_code: str = ''
    work_code: str = ''
    reserved: str = ''

    @classmethod
    def from_str(cls, line: str) -> 'Transaction':
        flds = line.split('\t') + ['', '', '', '', '', '']
        pin = flds[0]
        server_datetime = None
        try:
            server_datetime = datetime.datetime.strptime(flds[1], '%Y-%m-%d %H:%M:%S')
        except ValueError:
            pass
        checktype = flds[2]
        verifycode = flds[3]
        work_code = flds[4]
        reserved = flds[5]

        return cls(
            pin=pin,
            server_datetime=server_datetime,
            check_type=checktype,
            verify_code=verifycode,
            work_code=work_code,
            reserved=reserved,
            raw=line,
        )


_user_fieds_map = {
    'PIN': 'pin',
    'Passwd': 'password',
    'Card': 'card',
    'Grp': 'group',
    'TZ': 'tz',
    'Pri': 'privileges',
    'Verify': 'verify',
    'ViceCard': 'vice_card',
}


@da.dataclass(frozen=True)
class User:
    pin: str
    name: str
    password: str
    card: str
    group: str
    tz: str
    privileges: str
    raw: str
    verify: str = ''
    vice_card: str = ''

    @classmethod
    def from_str(cls, line: str) -> 'User':
        flds = _build_operlog_fields(line)
        return _fill_da_from_mapping(
            cls,
            _user_fieds_map,
            **{
                'raw': line,
                **flds
            }
        )


_DA = typing.TypeVar('_DA', covariant=True)


def _fill_da_from_mapping(
        model_type: typing.Type[_DA],
        mapping: typing.Mapping[str, str],
        **kwargs: typing.Any,
) -> _DA:
    model_data = {}
    fields_names = [f.name for f in da.fields(model_type)]
    for key, val in kwargs.items():
        normal_key = stringcase.snakecase(key)
        if key in mapping:
            normal_key = mapping[key]
        if normal_key in fields_names:
            model_data[normal_key] = val
    return model_type(**model_data)  # type: ignore


_fingerprint_fields_map = {
    'PIN': 'pin',
    'FID': 'fid',
    'TMP': 'tmp',
}


@da.dataclass(frozen=True)
class Fingerprint:
    pin: str
    fid: str
    tmp: str
    raw: str

    @classmethod
    def from_str(cls, line: str) -> 'Fingerprint':
        flds = _build_operlog_fields(line)
        return _fill_da_from_mapping(
            cls,
            _fingerprint_fields_map,
            **{
                'raw': line,
                **flds
            }
        )


@da.dataclass(frozen=True)
class Operation(ServerDatetimeMixin):
    object: str
    param_1: str
    param_2: str
    param_3: str
    raw: str
    operation: OperationEnum
    admin: str
    alarm: AlarmEnum = AlarmEnum.unknown

    @classmethod
    def from_str(cls, line: str) -> 'Operation':
        flds = line.split('\t')
        logtime = None
        try:
            logtime = datetime.datetime.strptime(flds[2], '%Y-%m-%d %H:%M:%S')
        except ValueError:
            pass
        object = flds[3]
        operation = OperationEnum(flds[0])
        return Operation(
            admin=flds[1],
            operation=operation,
            server_datetime=logtime,
            object=object,
            param_1=flds[4],
            param_2=flds[5],
            param_3=flds[6],
            raw=line,
            alarm=AlarmEnum(
                object) if operation == OperationEnum.alarm else AlarmEnum.unknown
        )


@da.dataclass(frozen=True)
class OperationLog:
    raw: str
    users: typing.List[User] = da.field(default_factory=list)
    fingerprints: typing.List[Fingerprint] = da.field(default_factory=list)
    operations: typing.List[Operation] = da.field(default_factory=list)

    @classmethod
    def from_str(cls, data: str) -> 'OperationLog':
        users = []
        fingerprints = []
        operations = []
        for line in data.split('\n'):
            ops = line.split(' ', 1)

            if ops[0] == 'OPLOG':
                operations.append(Operation.from_str(ops[1]))
            if ops[0] == 'USER':
                users.append(User.from_str(ops[1]))
            elif ops[0] == 'FP':
                fingerprints.append(Fingerprint.from_str(ops[1]))
        return cls(
            users=users,
            operations=operations,
            fingerprints=fingerprints,
            raw=data
        )


@da.dataclass(frozen=True)
class AttendanceLog:
    raw: str
    transactions: typing.List[Transaction] = da.field(default_factory=list)

    @classmethod
    def from_str(cls, data: str) -> 'AttendanceLog':
        transactions = []
        for line in data.split('\n'):
            transactions.append(Transaction.from_str(line))
        return cls(
            transactions=transactions,
            raw=data,
        )


@da.dataclass(frozen=True)
class AttendancePhotoLog(ServerDatetimeMixin):
    raw: str
    pin: str = ''
    is_uploadphoto: bool = False
    is_realupload: bool = False
    data: str = ''

    @classmethod
    def from_request_pin(cls, req_pin: str, body: str) -> 'AttendancePhotoLog':
        pin_split = req_pin.split('.')[0].split('-')  # type: typing.List[str]
        dt = pin_split[0]
        pin = ''
        if len(pin_split) == 2:  # Success Picture
            pin = pin_split[1]
        server_datetime = datetime.datetime.strptime(dt, '%Y-%m-%d %H:%M:%S')
        image_data = ''
        is_uploadphoto = False
        is_realupload = False

        if 'CMD=uploadphoto' in body:
            image_data = body.split('CMD=uploadphoto')[1][1:]
            is_uploadphoto = True
        if 'CMD=realupload' in body:
            image_data = body.split('CMD=realupload')[1][1:]
            is_realupload = True

        return cls(
            pin=pin,
            server_datetime=server_datetime,
            is_uploadphoto=is_uploadphoto,
            is_realupload=is_realupload,
            data=image_data,
            raw=req_pin + body,
        )


def _build_operlog_fields(ops_1: str) -> typing.Dict[str, typing.Any]:
    flds = {}
    for item in ops_1.split('\t'):
        index = item.find('=')
        if index > 0:
            flds[item[:index]] = item[index + 1:]

    return flds



import typing
from urllib.parse import urlencode
from urllib.request import Request

import dataclasses as da


@da.dataclass()
class DeviceRequestBuilder:
    sn: str
    iclock_host: str
    fw_version: str

    def request(
            self,
            path: str,
            query: typing.Dict[str, typing.Any] = None,
            body: bytes = b'',
    ) -> Request:
        query = {} if query is None else query
        query['SN'] = self.sn
        str_query = ''
        if query:
            str_query = urlencode(query)

        url = "{:s}{:s}?{:s}".format(self.iclock_host, path, str_query)

        if body:
            req = Request(url, body)
        else:
            req = Request(url)
        req.headers["Content-type"] = "text/plain"

        return req

    def getrequest(self, query: typing.Dict[str, typing.Any]) -> Request:
        return self.request(
            "/iclock/getrequest",
            query=query
        )

    def postcmdrequest(self, id, cmd, ret) -> Request:
        return self.request(
            "/iclock/devicecmd",
            body="ID={:s}&Return={:s}&CMD={:s}".format(id, ret, cmd).encode('ascii')
        )

    def cdatarequest(
            self,
            query: typing.Dict[str, typing.Any] = None,
            body: bytes = b''
    ) -> Request:
        return self.request(
            '/iclock/cdata',
            query,
            body
        )



import collections
import typing
from urllib.parse import ParseResult, parse_qs, urlparse
from urllib.request import Request

import dataclasses as da
import stringcase

from .models import AttendanceLog, AttendancePhotoLog, OperationLog, TableEnum


@da.dataclass(frozen=True)
class Info:
    fw_version: str = ''
    fp_count: int = 0
    transaction_count: str = ''
    user_count: int = 0
    main_time: str = ''
    max_finger_count: int = 0
    lock_fun_on: str = ''
    max_att_log_count: int = 0
    device_name: str = ''
    alg_ver: str = ''
    flash_size: str = ''
    free_flash_size: str = ''
    language: str = ''
    volume: str = ''
    dt_fmt: str = ''
    ip_address: str = ''
    is_tft: bool = False
    platform: str = ''
    brightness: str = ''
    backup_dev: str = ''
    oem_vendor: str = ''
    fp_version: str = ''


@da.dataclass(frozen=True)
class ZKRequest:
    sn: str
    push_version: str


@da.dataclass(frozen=True)
class GetRequest(ZKRequest):
    info: Info

    @staticmethod
    def from_req(req: Request):
        parsed_req = _ParsedRequest.from_req(req)
        (sn, pushver) = _extract_sn_version(parsed_req)
        return GetRequest(
            sn=sn,
            push_version=pushver,
            info=_fill_plain_info(parsed_req.params.get('INFO', '')),
        )


@da.dataclass(frozen=True)
class _ParsedRequest:
    req: Request
    method: str
    parseresult: ParseResult
    params: typing.Dict[str, typing.Any]
    body: bytes
    headers: typing.Dict[str, str]

    @classmethod
    def from_req(cls, req: Request) -> '_ParsedRequest':
        parseresult = urlparse(req.get_full_url())
        return cls(
            req=req,
            method=req.get_method(),
            headers=req.headers,
            body=req.data or b'',
            parseresult=parseresult,
            params={k: v if len(v) > 1 else v[0] for k, v in
                    parse_qs(parseresult.query).items()},
        )


@da.dataclass(frozen=True)
class CdataRequest(ZKRequest):
    method: str
    pin: str = ''
    save: bool = False
    body: str = ''
    stamp: str = ''
    operation_stamp: str = ''
    table: TableEnum = TableEnum.unknown
    attendance_log: typing.Optional[AttendanceLog] = None
    operation_log: typing.Optional[OperationLog] = None
    attendance_photo_log: typing.Optional[AttendancePhotoLog] = None

    @staticmethod
    def from_req(req: Request) -> 'CdataRequest':
        parsed_req = _ParsedRequest.from_req(req)

        (sn, pushver) = _extract_sn_version(parsed_req)
        method = parsed_req.method
        action = parsed_req.params.get('action', '')
        if action:
            return CdataRequest(
                sn=sn,
                push_version=pushver,
                method=method,
            )
        if method == 'GET':
            pin = _from_maps('PIN', 0, parsed_req.params)
            save = _from_maps('save', '', parsed_req.params) in ['1', 'Y', 'y', 'yes',
                                                                 'YES']
            return CdataRequest(
                sn=sn,
                push_version=pushver,
                method=method,
                pin=pin,
                save=save,
            )
        if method == 'POST':
            stamp = _from_maps('Stamp', '', parsed_req.params)
            operation_stamp = _from_maps('OpStamp', '', parsed_req.params)
            table = TableEnum(_from_maps('table', None, parsed_req.params))
            body = ''
            try:
                body = parsed_req.body.decode('ascii')
            except UnicodeDecodeError:
                try:
                    body = parsed_req.body.decode('gb18030')
                except UnicodeDecodeError:
                    pass

            operlog = att_log = att_photo = None
            if table == TableEnum.operlog:
                operlog = OperationLog.from_str(body)

            if table == TableEnum.attlog:
                att_log = AttendanceLog.from_str(body)

            if table == TableEnum.attphoto:
                att_photo = AttendancePhotoLog.from_request_pin(
                    _from_maps('PIN', '', parsed_req.params),
                    body,
                )

            return CdataRequest(
                sn=sn,
                push_version=pushver,
                method=method,
                table=table,
                stamp=stamp,
                operation_stamp=operation_stamp,
                body=body,
                attendance_log=att_log,
                attendance_photo_log=att_photo,
                operation_log=operlog
            )

        return CdataRequest(
            sn=sn,
            push_version=pushver,
            method=method,
        )


def _from_maps(key: str, defaut: typing.Any,
               *args: typing.Mapping[str, typing.Any]) -> typing.Any:
    for d in args:
        if key in d:
            return d[key]
    return defaut


def _extract_sn_version(req: _ParsedRequest) -> typing.Tuple[str, str]:
    pushver = req.params.get('pushver', 0.0)
    sn = _from_maps('SN', '', req.params)

    if not sn:
        sn = req.req.get_full_url()
        sn = (sn + 'SN=').split('SN=')[1].split('&')[0]
        if sn == '':
            sn = 'UNKNOWN'
    return sn, pushver


_info_map = collections.OrderedDict({
    'FWVersion': 'fw_version',
    'FPCount': 'fp_count',
    'VOLUME': 'volume',
    'IPAddress': 'ip_address',
    'IsTFT': 'is_tft',
    'OEMVendor': 'oem_vendor',
    'FPVersion': ''
}
)

_info_int_fields = frozenset(
    ['fp_count', 'transaction_count', 'user_count', 'max_finger_count',
     'max_att_log_count', ])


def _fill_plain_info(info: str) -> Info:
    if info:
        splitted_info = info.split(',')

        if len(splitted_info) >= 6:
            info = 'FWVersion={}\tUserCount={}\tFPCount={}' \
                   '\tTransactionCount={}\tIPAddress={}\tFPVersion={}\t' \
                .format(*splitted_info[:6])
        elif len(splitted_info) == 5:
            info = 'FWVersion={}\tUserCount={}\tFPCount={}' \
                   '\tTransactionCount={}\tIPAddress={}\t' \
                .format(*splitted_info)
        elif len(splitted_info) == 4:
            info = 'FWVersion={}\tUserCount={}\tFPCount={}' \
                   '\tTransactionCount={}\t' \
                .format(*splitted_info)

    return _fill_info(info)


def _fill_info(info: str) -> Info:
    pd = _set_value_dict(info)
    info_fileds = frozenset([f.name for f in da.fields(Info)])
    info_data = {}
    for key in pd.keys():
        normal_key = key
        if key[0] == '~':
            normal_key = key[1:]

        if normal_key in _info_map:
            normal_key = _info_map[normal_key]
        else:
            normal_key = stringcase.snakecase(normal_key)

        if normal_key in info_fileds:
            if normal_key == 'platform' and '_TFT' in pd[key]:
                info_data['is_tft'] = True
            value = pd[key]
            # ints
            if normal_key in _info_int_fields:
                value = 0
                try:
                    value = int(pd[key])
                except ValueError:
                    pass

            if normal_key == 'max_att_log_count':
                value *= 10000
            if normal_key == 'max_finger_count':
                value *= 100
            info_data[normal_key] = value

    return Info(
        **info_data  # type: ignore
    )


def _set_value_dict(data: str) -> typing.Mapping[str, typing.Any]:
    d = {}
    for line in data.split('\t'):
        if line:
            v = line.split('\r')[0]
        else:
            v = line
        nv = v.split('=', 1)
        if len(nv) > 1:
            try:
                v = str(nv[1])
                d[nv[0]] = v
            except ValueError:
                pass
    return d



import datetime
import typing
import unittest

import dataclasses as da

from iclockhelper.models import (
    AlarmEnum,
    Fingerprint,
    Operation,
    OperationEnum,
    TableEnum,
    Transaction,
    User
)
from iclockhelper.requests import CdataRequest, GetRequest

from .common import DeviceRequestBuilder

_SN = 'SN_C_50'
_ICLOCK_HOST = 'http://localhost'
_FW_VERSION = '2.4.0'
_LANGUAGE = 69
_USER_COUNT = 3
_FP_COUNT = 2
_TRANS_COUNT = 10
_STAMP = '11122'
_OP_STAMP = '9999'


def _create_many_trans_body(transactions: typing.List[Transaction]) -> str:
    return "\n".join([
        "{:s}\t{:s}\t{:s}\t{:s}\t{:s}\t{:s}".format(
            trans.pin,
            trans.server_datetime.strftime('%Y-%m-%d %H:%M:%S')
            if trans.server_datetime else '',
            trans.check_type,
            trans.verify_code,
            trans.work_code,
            trans.reserved,
        ) for trans in
        transactions])


def _create_many_user_body(users: typing.List[User]) -> str:
    return "\n".join(
        [
            "USER PIN={:s}\tName={:s}\tPri={:s}\tPasswd={:s}"
            "\tCard={:s}\tGrp={:s}\tTZ={:s}\tVerify={:s}\tViceCard={:s}".format(
                u.pin,
                u.name,
                u.privileges,
                u.password,
                u.card,
                u.group,
                u.tz,
                u.verify,
                u.vice_card,
            ) for u in users])


def _create_many_fingerprint_body(fingerprints: typing.List[Fingerprint]) -> str:
    return "\n".join(
        ["FP PIN={:s}\tFID={:s}\tValid={:s}\tTMP={:s}".format(f.pin, f.fid, '1', f.tmp)
         for f in fingerprints])


def _create_many_operation_body(opers: typing.List[Operation]) -> str:
    return "\n".join(
        ["OPLOG {:s}\t{:s}\t{:s}\t{:s}\t{:s}\t{:s}\t{:s}".format(
            o.operation.value,
            o.admin,
            o.server_datetime.strftime('%Y-%m-%d %H:%M:%S')
            if o.server_datetime else '',
            o.object,
            o.param_1,
            o.param_2,
            o.param_3,
        ) for o in opers])


class ModelC50Test(unittest.TestCase):
    def setUp(self) -> None:
        self.req_builder = DeviceRequestBuilder(
            sn=_SN,
            iclock_host=_ICLOCK_HOST,
            fw_version=_FW_VERSION,
        )

    def test_cdata_all_options(self):
        base_datetime = datetime.datetime(
            year=2000,
            month=1,
            day=1,
            hour=1,
            minute=1,
            second=0)

        trans = [
            Transaction(
                pin='pin1',
                server_datetime=base_datetime + datetime.timedelta(seconds=5),
                raw=''
            ),
            Transaction(
                pin='pin1',
                server_datetime=base_datetime + datetime.timedelta(seconds=10),
                raw=''
            )
        ]
        body = _create_many_trans_body(trans)
        req = self.req_builder.cdatarequest(
            query={
                'options': 'all',
                'pushver': _FW_VERSION,
                'language': _LANGUAGE,
                'PushOptionsFlag': 1
            },
            body=body.encode('ascii')
        )
        cdata_req = CdataRequest.from_req(req)
        self.assertEqual(_SN, cdata_req.sn)
        self.assertTrue(_FW_VERSION, cdata_req.push_version)

    def test_cdata_operlog_oplog(self):
        base_datetime = datetime.datetime(year=2000, month=1, day=1, hour=1, minute=1,
                                          second=0)

        operations = [
            Operation(
                object='51',
                param_1='p11',
                param_2='p12',
                param_3='p13',
                raw='',
                admin='a1',
                operation=OperationEnum.alarm,
                alarm=AlarmEnum.door_open_detected,
                server_datetime=base_datetime + datetime.timedelta(seconds=5),
            ),
            Operation(
                object='o2',
                param_1='p21',
                param_2='p22',
                param_3='p23',
                raw='',
                admin='a1',
                operation=OperationEnum.enter_the_menu,
                server_datetime=base_datetime + datetime.timedelta(seconds=5),
            )
        ]
        body = _create_many_operation_body(operations)
        req = self.req_builder.cdatarequest(
            query={
                'table': TableEnum.operlog.value,
                'OpStamp': _OP_STAMP
            },
            body=body.encode('ascii'),
        )
        cdata_req = CdataRequest.from_req(req)
        self.assertEqual(_SN, cdata_req.sn)
        self.assertTrue(_OP_STAMP, cdata_req.operation_stamp)
        self.assertTrue(TableEnum.operlog, cdata_req.table)
        self.assertIsNotNone(cdata_req.operation_log)
        self.assertEqual(len(operations), len(cdata_req.operation_log.operations))
        for i in range(len(operations)):
            expected = da.asdict(operations[i])
            actual = da.asdict(cdata_req.operation_log.operations[i])
            del expected['raw']
            del actual['raw']
            self.assertEqual(
                expected,
                actual,
            )

    def test_cdata_operlog_user(self):
        operations = [
            User(
                pin='pin1',
                name='name1',
                password='pass1',
                card='card1',
                group='group1',
                tz='tz1',
                privileges='privileges1',
                verify='verify1',
                vice_card='vise_card1',
                raw=''
            ),
            User(
                pin='pin2',
                name='name2',
                password='pass2',
                card='card2',
                group='group2',
                tz='tz2',
                privileges='privileges2',
                verify='verify2',
                vice_card='vise_card2',
                raw=''
            )
        ]
        body = _create_many_user_body(operations)
        req = self.req_builder.cdatarequest(
            query={
                'table': TableEnum.operlog.value,
                'OpStamp': '9999'
            },
            body=body.encode('ascii'),
        )
        cdata_req = CdataRequest.from_req(req)
        self.assertEqual(_SN, cdata_req.sn)
        self.assertTrue(_OP_STAMP, cdata_req.operation_stamp)
        self.assertTrue(TableEnum.operlog, cdata_req.table)
        self.assertIsNotNone(cdata_req.operation_log)
        self.assertEqual(len(operations), len(cdata_req.operation_log.users))
        for i in range(len(operations)):
            expected = da.asdict(operations[i])
            actual = da.asdict(cdata_req.operation_log.users[i])
            del expected['raw']
            del actual['raw']
            self.assertEqual(
                expected,
                actual,
            )

    def test_cdata_operlog_fp(self):
        operations = [
            Fingerprint(
                pin='pin1',
                fid='fid1',
                tmp='tmp1',
                raw=''
            ),
            Fingerprint(
                pin='pin2',
                fid='fid2',
                tmp='tmp2',
                raw=''
            )
        ]
        body = _create_many_fingerprint_body(operations)
        req = self.req_builder.cdatarequest(
            query={
                'table': TableEnum.operlog.value,
                'OpStamp': '9999'
            },
            body=body.encode('ascii'),
        )
        cdata_req = CdataRequest.from_req(req)
        self.assertEqual(_SN, cdata_req.sn)
        self.assertTrue(_OP_STAMP, cdata_req.operation_stamp)
        self.assertTrue(TableEnum.operlog, cdata_req.table)
        self.assertIsNotNone(cdata_req.operation_log)
        self.assertEqual(len(operations), len(cdata_req.operation_log.fingerprints))
        for i in range(len(operations)):
            expected = da.asdict(operations[i])
            actual = da.asdict(cdata_req.operation_log.fingerprints[i])
            del expected['raw']
            del actual['raw']
            self.assertEqual(
                expected,
                actual,
            )

    def test_cdata_attlog(self):
        base_datetime = datetime.datetime(year=2000, month=1, day=1, hour=1, minute=1,
                                          second=0)

        operations = [
            Transaction(
                pin='pin1',
                server_datetime=base_datetime + datetime.timedelta(seconds=5),
                check_type='ct1',
                verify_code='vc1',
                work_code='wc1',
                reserved='re1',
                raw=''
            ),
            Transaction(
                pin='pin1',
                server_datetime=base_datetime + datetime.timedelta(seconds=10),
                check_type='ct2',
                verify_code='vc2',
                work_code='wc2',
                reserved='re2',
                raw=''
            )
        ]
        body = _create_many_trans_body(operations)
        req = self.req_builder.cdatarequest(
            query={'table': TableEnum.attlog.value, 'Stamp': '9999'},
            body=body.encode('ascii'),
        )
        cdata_req = CdataRequest.from_req(req)
        self.assertEqual(_SN, cdata_req.sn)
        self.assertTrue(_STAMP, cdata_req.stamp)
        self.assertTrue(TableEnum.attlog, cdata_req.table)
        self.assertIsNotNone(cdata_req.attendance_log)
        self.assertEqual(len(operations), len(cdata_req.attendance_log.transactions))
        for i in range(len(operations)):
            expected = da.asdict(operations[i])
            actual = da.asdict(cdata_req.attendance_log.transactions[i])
            del expected['raw']
            del actual['raw']
            self.assertEqual(
                expected,
                actual,
            )

    def test_getreq(self):
        req = self.req_builder.getrequest(
            query={
                "INFO": "{:s},{:d},{:d},{:d},{:s}".format(
                    _FW_VERSION,
                    _USER_COUNT,
                    _FP_COUNT,
                    _TRANS_COUNT,
                    _ICLOCK_HOST)
            }
        )
        get_req = GetRequest.from_req(req)

        self.assertEqual(_SN, get_req.sn)
        self.assertEqual(_FW_VERSION, get_req.info.fw_version)
        self.assertEqual(_USER_COUNT, get_req.info.user_count)
        self.assertEqual(_FP_COUNT, get_req.info.fp_count)
        self.assertEqual(_TRANS_COUNT, get_req.info.transaction_count)

    def test_devpostreq(self):
        self.assertTrue(True)

    def test_fdatareq(self):
        self.assertTrue(True)



import datetime

import pytz as pytz

from iclockhelper.models import ServerDatetimeMixin


def test_server_datetime_mixin():
    base_datetime = datetime.datetime(
        year=2000,
        month=1,
        day=1,
        hour=1,
        minute=1,
        second=0,
        tzinfo=pytz.timezone("UTC"),
    )
    gmt = pytz.timezone("GMT")
    actual_time = ServerDatetimeMixin(server_datetime=base_datetime).correct_datetime(
        gmt
    )
    for f in ['year', 'month', 'day', 'hour', 'minute', 'second']:
        assert getattr(base_datetime, f) == getattr(actual_time, f)

    assert actual_time.tzinfo == gmt
